# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
A script to convert the standard names information from the provided XML
file into a Python dictionary format.

Takes two arguments: the first is the XML file to process and the second
is the name of the file to write the Python dictionary file into.

By default, Iris will use the source XML file:
    etc/cf-standard-name-table.xml
as obtained from:
    http://cf-pcmdi.llnl.gov/documents/cf-standard-names

"""

import argparse
import pprint
import xml.etree.ElementTree as ET


STD_VALUES_FILE_TEMPLATE = '''
# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
This file contains a dictionary of standard value names that are mapped
to another dictionary of other standard name attributes. Currently only
the `canonical_unit` exists in these attribute dictionaries.

This file is automatically generated. Do not edit this file by hand.

The file will be generated during a standard build/installation::

    python setup.py build
    python setup.py install

Also, the file can be re-generated in the source distribution via::

    python setup.py std_names

Or for more control (e.g. to use an alternative XML file) via::

    python tools/generate_std_names.py XML_FILE MODULE_FILE

"""

STD_NAMES = '''.lstrip()


def process_name_table(tree, element_name, *child_elements):
    """
    Yields a series of dictionaries with the key being the id of the entry element and the value containing
    another dictionary mapping other attributes of the standard name to their values, e.g. units, description, grib value etc.
    """
    for elem in tree.iterfind(element_name):
        sub_section = {}

        for child_elem in child_elements:
            found_elem = elem.find(child_elem)
            sub_section[child_elem] = found_elem.text if found_elem is not None else None

        yield {elem.get("id") : sub_section}


def to_dict(infile, outfile):
    values = {}
    aliases = {}

    tree = ET.parse(infile)

    for section in process_name_table(tree, 'entry', 'canonical_units'):
        values.update(section)

    for section in process_name_table(tree, 'alias', 'entry_id'):
        aliases.update(section)

    for key, valued in aliases.items():
        values.update({
                key : {'canonical_units' : values.get(valued['entry_id']).get('canonical_units')}
            })

    outfile.write(STD_VALUES_FILE_TEMPLATE + pprint.pformat(values))


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description='Create Python code from CF standard name XML.')
    parser.add_argument('input', metavar='INPUT',
                        help='Path to CF standard name XML')
    parser.add_argument('output', metavar='OUTPUT',
                        help='Path to resulting Python code')
    args = parser.parse_args()

    encoding = {'encoding': 'utf-8'}

    with open(args.input, 'r', **encoding) as in_fh:
        with open(args.output, 'w', **encoding) as out_fh:
            to_dict(in_fh, out_fh)



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the :class:`iris.fileformat.ff.FF2PP` class."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import collections
import contextlib
from unittest import mock

import numpy as np

from iris.exceptions import NotYetImplementedError
import iris.fileformats._ff as ff
import iris.fileformats.pp as pp
from iris.fileformats._ff import FF2PP


# PP-field: LBPACK N1 values.
_UNPACKED = 0
_WGDOS = 1
_CRAY = 2

# PP-field: LBUSER(1) values.
_REAL = 1
_INTEGER = 2


_DummyField = collections.namedtuple(
    "_DummyField", "lbext lblrec lbnrec raw_lbpack " "lbuser boundary_packing"
)
_DummyFieldWithSize = collections.namedtuple(
    "_DummyFieldWithSize",
    "lbext lblrec lbnrec raw_lbpack " "lbuser boundary_packing " "lbnpt lbrow",
)
_DummyBoundaryPacking = collections.namedtuple(
    "_DummyBoundaryPacking", "x_halo y_halo rim_width"
)


class Test____iter__(tests.IrisTest):
    @mock.patch("iris.fileformats._ff.FFHeader")
    def test_call_structure(self, _FFHeader):
        # Check that the iter method calls the two necessary utility
        # functions
        extract_result = mock.Mock()
        interpret_patch = mock.patch(
            "iris.fileformats.pp._interpret_fields",
            autospec=True,
            return_value=iter([]),
        )
        extract_patch = mock.patch(
            "iris.fileformats._ff.FF2PP._extract_field",
            autospec=True,
            return_value=extract_result,
        )

        FF2PP_instance = ff.FF2PP("mock")
        with interpret_patch as interpret, extract_patch as extract:
            list(iter(FF2PP_instance))

        interpret.assert_called_once_with(extract_result)
        extract.assert_called_once_with(FF2PP_instance)


class Test__extract_field__LBC_format(tests.IrisTest):
    @contextlib.contextmanager
    def mock_for_extract_field(self, fields, x=None, y=None):
        """
        A context manager to ensure FF2PP._extract_field gets a field
        instance looking like the next one in the "fields" iterable from
        the "make_pp_field" call.

        """
        with mock.patch("iris.fileformats._ff.FFHeader"):
            ff2pp = ff.FF2PP("mock")
        ff2pp._ff_header.lookup_table = [0, 0, len(fields)]
        # Fake level constants, with shape specifying just one model-level.
        ff2pp._ff_header.level_dependent_constants = np.zeros(1)
        grid = mock.Mock()
        grid.vectors = mock.Mock(return_value=(x, y))
        ff2pp._ff_header.grid = mock.Mock(return_value=grid)

        open_func = "builtins.open"
        with mock.patch(
            "iris.fileformats._ff._parse_binary_stream", return_value=[0]
        ), mock.patch(open_func), mock.patch(
            "struct.unpack_from", return_value=[4]
        ), mock.patch(
            "iris.fileformats.pp.make_pp_field", side_effect=fields
        ), mock.patch(
            "iris.fileformats._ff.FF2PP._payload", return_value=(0, 0)
        ):
            yield ff2pp

    def _mock_lbc(self, **kwargs):
        """Return a Mock object representing an LBC field."""
        # Default kwargs for a valid LBC field mapping just 1 model-level.
        field_kwargs = dict(lbtim=0, lblev=7777, lbvc=0, lbhem=101)
        # Apply provided args (replacing any defaults if specified).
        field_kwargs.update(kwargs)
        # Return a mock with just those properties pre-defined.
        return mock.Mock(**field_kwargs)

    def test_LBC_header(self):
        bzx, bzy = -10, 15
        # stash m01s00i001
        lbuser = [None, None, 121416, 1, None, None, 1]
        field = self._mock_lbc(
            lbegin=0,
            lbrow=10,
            lbnpt=12,
            bdx=1,
            bdy=1,
            bzx=bzx,
            bzy=bzy,
            lbuser=lbuser,
        )
        with self.mock_for_extract_field([field]) as ff2pp:
            ff2pp._ff_header.dataset_type = 5
            result = list(ff2pp._extract_field())

        self.assertEqual([field], result)
        self.assertEqual(field.lbrow, 10 + 14 * 2)
        self.assertEqual(field.lbnpt, 12 + 16 * 2)

        name_mapping_dict = dict(
            rim_width=slice(4, 6), y_halo=slice(2, 4), x_halo=slice(0, 2)
        )
        boundary_packing = pp.SplittableInt(121416, name_mapping_dict)

        self.assertEqual(field.boundary_packing, boundary_packing)
        self.assertEqual(field.bzy, bzy - boundary_packing.y_halo * field.bdy)
        self.assertEqual(field.bzx, bzx - boundary_packing.x_halo * field.bdx)

    def check_non_trivial_coordinate_warning(self, field):
        field.lbegin = 0
        field.lbrow = 10
        field.lbnpt = 12
        # stash m01s31i020
        field.lbuser = [None, None, 121416, 20, None, None, 1]
        orig_bdx, orig_bdy = field.bdx, field.bdy

        x = np.array([1, 2, 6])
        y = np.array([1, 2, 6])
        with self.mock_for_extract_field([field], x, y) as ff2pp:
            ff2pp._ff_header.dataset_type = 5
            with mock.patch("warnings.warn") as warn:
                list(ff2pp._extract_field())

        # Check the values are unchanged.
        self.assertEqual(field.bdy, orig_bdy)
        self.assertEqual(field.bdx, orig_bdx)

        # Check a warning was raised with a suitable message.
        warn_error_tmplt = "Unexpected warning message: {}"
        non_trivial_coord_warn_msg = warn.call_args[0][0]
        msg = (
            "The x or y coordinates of your boundary condition field may "
            "be incorrect, not having taken into account the boundary "
            "size."
        )
        self.assertTrue(
            non_trivial_coord_warn_msg.startswith(msg),
            warn_error_tmplt.format(non_trivial_coord_warn_msg),
        )

    def test_LBC_header_non_trivial_coords_both(self):
        # Check a warning is raised when both bdx and bdy are bad.
        field = self._mock_lbc(bdx=0, bdy=0, bzx=10, bzy=10)
        self.check_non_trivial_coordinate_warning(field)

        field.bdy = field.bdx = field.bmdi
        self.check_non_trivial_coordinate_warning(field)

    def test_LBC_header_non_trivial_coords_x(self):
        # Check a warning is raised when bdx is bad.
        field = self._mock_lbc(bdx=0, bdy=10, bzx=10, bzy=10)
        self.check_non_trivial_coordinate_warning(field)

        field.bdx = field.bmdi
        self.check_non_trivial_coordinate_warning(field)

    def test_LBC_header_non_trivial_coords_y(self):
        # Check a warning is raised when bdy is bad.
        field = self._mock_lbc(bdx=10, bdy=0, bzx=10, bzy=10)
        self.check_non_trivial_coordinate_warning(field)

        field.bdy = field.bmdi
        self.check_non_trivial_coordinate_warning(field)

    def test_negative_bdy(self):
        # Check a warning is raised when bdy is negative,
        # we don't yet know what "north" means in this case.
        field = self._mock_lbc(
            bdx=10,
            bdy=-10,
            bzx=10,
            bzy=10,
            lbegin=0,
            lbuser=[0, 0, 121416, 0, None, None, 0],
            lbrow=10,
            lbnpt=12,
        )
        with self.mock_for_extract_field([field]) as ff2pp:
            ff2pp._ff_header.dataset_type = 5
            with mock.patch("warnings.warn") as warn:
                list(ff2pp._extract_field())
        msg = "The LBC has a bdy less than 0."
        self.assertTrue(
            warn.call_args[0][0].startswith(msg),
            "Northwards bdy warning not correctly raised.",
        )


class Test__payload(tests.IrisTest):
    def setUp(self):
        # Create a mock LBC type PPField.
        self.mock_field = mock.Mock()
        field = self.mock_field
        field.raw_lbpack = _UNPACKED
        field.lbuser = [_REAL]
        field.lblrec = 777
        field.lbext = 222
        field.lbnrec = 50
        field.boundary_packing = None

    def _test(
        self, mock_field, expected_depth, expected_dtype, word_depth=None
    ):
        with mock.patch("iris.fileformats._ff.FFHeader", return_value=None):
            kwargs = {}
            if word_depth is not None:
                kwargs["word_depth"] = word_depth
            ff2pp = FF2PP("dummy_filename", **kwargs)
            data_depth, data_dtype = ff2pp._payload(mock_field)
            self.assertEqual(data_depth, expected_depth)
            self.assertEqual(data_dtype, expected_dtype)

    def test_unpacked_real(self):
        mock_field = _DummyField(
            lbext=0,
            lblrec=100,
            lbnrec=-1,
            raw_lbpack=_UNPACKED,
            lbuser=[_REAL],
            boundary_packing=None,
        )
        self._test(mock_field, 800, ">f8")

    def test_unpacked_real_ext(self):
        mock_field = _DummyField(
            lbext=5,
            lblrec=100,
            lbnrec=-1,
            raw_lbpack=_UNPACKED,
            lbuser=[_REAL],
            boundary_packing=None,
        )
        self._test(mock_field, 760, ">f8")

    def test_unpacked_integer(self):
        mock_field = _DummyField(
            lbext=0,
            lblrec=200,
            lbnrec=-1,
            raw_lbpack=_UNPACKED,
            lbuser=[_INTEGER],
            boundary_packing=None,
        )
        self._test(mock_field, 1600, ">i8")

    def test_unpacked_integer_ext(self):
        mock_field = _DummyField(
            lbext=10,
            lblrec=200,
            lbnrec=-1,
            raw_lbpack=_UNPACKED,
            lbuser=[_INTEGER],
            boundary_packing=None,
        )
        self._test(mock_field, 1520, ">i8")

    def test_unpacked_real_ext_different_word_depth(self):
        mock_field = _DummyField(
            lbext=5,
            lblrec=100,
            lbnrec=-1,
            raw_lbpack=_UNPACKED,
            lbuser=[_REAL],
            boundary_packing=None,
        )
        self._test(mock_field, 380, ">f4", word_depth=4)

    def test_wgdos_real(self):
        mock_field = _DummyField(
            lbext=0,
            lblrec=-1,
            lbnrec=100,
            raw_lbpack=_WGDOS,
            lbuser=[_REAL],
            boundary_packing=None,
        )
        self._test(mock_field, 800, ">f4")

    def test_wgdos_real_ext(self):
        mock_field = _DummyField(
            lbext=5,
            lblrec=-1,
            lbnrec=100,
            raw_lbpack=_WGDOS,
            lbuser=[_REAL],
            boundary_packing=None,
        )
        self._test(mock_field, 800, ">f4")

    def test_wgdos_integer(self):
        mock_field = _DummyField(
            lbext=0,
            lblrec=-1,
            lbnrec=200,
            raw_lbpack=_WGDOS,
            lbuser=[_INTEGER],
            boundary_packing=None,
        )
        self._test(mock_field, 1600, ">i4")

    def test_wgdos_integer_ext(self):
        mock_field = _DummyField(
            lbext=10,
            lblrec=-1,
            lbnrec=200,
            raw_lbpack=_WGDOS,
            lbuser=[_INTEGER],
            boundary_packing=None,
        )
        self._test(mock_field, 1600, ">i4")

    def test_cray_real(self):
        mock_field = _DummyField(
            lbext=0,
            lblrec=100,
            lbnrec=-1,
            raw_lbpack=_CRAY,
            lbuser=[_REAL],
            boundary_packing=None,
        )
        self._test(mock_field, 400, ">f4")

    def test_cray_real_ext(self):
        mock_field = _DummyField(
            lbext=5,
            lblrec=100,
            lbnrec=-1,
            raw_lbpack=_CRAY,
            lbuser=[_REAL],
            boundary_packing=None,
        )
        self._test(mock_field, 380, ">f4")

    def test_cray_integer(self):
        mock_field = _DummyField(
            lbext=0,
            lblrec=200,
            lbnrec=-1,
            raw_lbpack=_CRAY,
            lbuser=[_INTEGER],
            boundary_packing=None,
        )
        self._test(mock_field, 800, ">i4")

    def test_cray_integer_ext(self):
        mock_field = _DummyField(
            lbext=10,
            lblrec=200,
            lbnrec=-1,
            raw_lbpack=_CRAY,
            lbuser=[_INTEGER],
            boundary_packing=None,
        )
        self._test(mock_field, 760, ">i4")

    def test_lbpack_unsupported(self):
        mock_field = _DummyField(
            lbext=10,
            lblrec=200,
            lbnrec=-1,
            raw_lbpack=1239,
            lbuser=[_INTEGER],
            boundary_packing=None,
        )
        with self.assertRaisesRegex(
            NotYetImplementedError,
            "PP fields with LBPACK of 1239 are not supported.",
        ):
            self._test(mock_field, None, None)

    def test_lbc_unpacked(self):
        boundary_packing = _DummyBoundaryPacking(
            x_halo=11, y_halo=7, rim_width=3
        )
        mock_field = _DummyFieldWithSize(
            lbext=10,
            lblrec=200,
            lbnrec=-1,
            raw_lbpack=_UNPACKED,
            lbuser=[_REAL],
            boundary_packing=boundary_packing,
            lbnpt=47,
            lbrow=34,
        )
        self._test(mock_field, ((47 * 34) - (19 * 14)) * 8, ">f8")

    def test_lbc_wgdos_unsupported(self):
        mock_field = _DummyField(
            lbext=5,
            lblrec=-1,
            lbnrec=100,
            raw_lbpack=_WGDOS,
            lbuser=[_REAL],
            # Anything not None will do here.
            boundary_packing=0,
        )
        with self.assertRaisesRegex(
            ValueError, "packed LBC data is not supported"
        ):
            self._test(mock_field, None, None)

    def test_lbc_cray(self):
        boundary_packing = _DummyBoundaryPacking(
            x_halo=11, y_halo=7, rim_width=3
        )
        mock_field = _DummyFieldWithSize(
            lbext=10,
            lblrec=200,
            lbnrec=-1,
            raw_lbpack=_CRAY,
            lbuser=[_REAL],
            boundary_packing=boundary_packing,
            lbnpt=47,
            lbrow=34,
        )
        self._test(mock_field, ((47 * 34) - (19 * 14)) * 4, ">f4")


class Test__det_border(tests.IrisTest):
    def setUp(self):
        _FFH_patch = mock.patch("iris.fileformats._ff.FFHeader")
        _FFH_patch.start()
        self.addCleanup(_FFH_patch.stop)

    def test_unequal_spacing_eitherside(self):
        # Ensure that we do not interpret the case where there is not the same
        # spacing on the lower edge as the upper edge.
        ff2pp = FF2PP("dummy")
        field_x = np.array([1, 2, 10])

        msg = (
            "The x or y coordinates of your boundary condition field may "
            "be incorrect, not having taken into account the boundary "
            "size."
        )

        with mock.patch("warnings.warn") as warn:
            result = ff2pp._det_border(field_x, None)
        warn.assert_called_with(msg)
        self.assertIs(result, field_x)

    def test_increasing_field_values(self):
        # Field where its values a increasing.
        ff2pp = FF2PP("dummy")
        field_x = np.array([1, 2, 3])
        com = np.array([0, 1, 2, 3, 4])
        result = ff2pp._det_border(field_x, 1)
        self.assertArrayEqual(result, com)

    def test_decreasing_field_values(self):
        # Field where its values a decreasing.
        ff2pp = FF2PP("dummy")
        field_x = np.array([3, 2, 1])
        com = np.array([4, 3, 2, 1, 0])
        result = ff2pp._det_border(field_x, 1)
        self.assertArrayEqual(result, com)


class Test__adjust_field_for_lbc(tests.IrisTest):
    def setUp(self):
        # Patch FFHeader to produce a mock header instead of opening a file.
        self.mock_ff_header = mock.Mock()
        self.mock_ff_header.dataset_type = 5
        self.mock_ff = self.patch(
            "iris.fileformats._ff.FFHeader", return_value=self.mock_ff_header
        )

        # Create a mock LBC type PPField.
        self.mock_field = mock.Mock()
        field = self.mock_field
        field.lbtim = 0
        field.lblev = 7777
        field.lbvc = 0
        field.lbnpt = 1001
        field.lbrow = 2001
        field.lbuser = (None, None, 80504)
        field.lbpack = pp.SplittableInt(0)
        field.boundary_packing = None
        field.bdx = 1.0
        field.bzx = 0.0
        field.bdy = 1.0
        field.bzy = 0.0

    def test__basic(self):
        ff2pp = FF2PP("dummy_filename")
        field = self.mock_field
        ff2pp._adjust_field_for_lbc(field)
        self.assertEqual(field.lbtim, 11)
        self.assertEqual(field.lbvc, 65)
        self.assertEqual(field.boundary_packing.rim_width, 8)
        self.assertEqual(field.boundary_packing.y_halo, 5)
        self.assertEqual(field.boundary_packing.x_halo, 4)
        self.assertEqual(field.lbnpt, 1009)
        self.assertEqual(field.lbrow, 2011)

    def test__bad_lbtim(self):
        self.mock_field.lbtim = 717
        ff2pp = FF2PP("dummy_filename")
        with self.assertRaisesRegex(
            ValueError, "LBTIM of 717, expected only 0 or 11"
        ):
            ff2pp._adjust_field_for_lbc(self.mock_field)

    def test__bad_lbvc(self):
        self.mock_field.lbvc = 312
        ff2pp = FF2PP("dummy_filename")
        with self.assertRaisesRegex(
            ValueError, "LBVC of 312, expected only 0 or 65"
        ):
            ff2pp._adjust_field_for_lbc(self.mock_field)


class Test__fields_over_all_levels(tests.IrisTest):
    def setUp(self):
        # Patch FFHeader to produce a mock header instead of opening a file.
        self.mock_ff_header = mock.Mock()
        self.mock_ff_header.dataset_type = 5

        # Fake the level constants to look like 3 model levels.
        self.n_all_levels = 3
        self.mock_ff_header.level_dependent_constants = np.zeros(
            (self.n_all_levels)
        )
        self.mock_ff = self.patch(
            "iris.fileformats._ff.FFHeader", return_value=self.mock_ff_header
        )

        # Create a simple mock for a test field.
        self.mock_field = mock.Mock()
        field = self.mock_field
        field.lbhem = 103
        self.original_lblev = mock.sentinel.untouched_lbev
        field.lblev = self.original_lblev

    def _check_expected_levels(self, results, n_levels):
        if n_levels == 0:
            self.assertEqual(len(results), 1)
            self.assertEqual(results[0].lblev, self.original_lblev)
        else:
            self.assertEqual(len(results), n_levels)
            self.assertEqual(
                [fld.lblev for fld in results], list(range(n_levels))
            )

    def test__is_lbc(self):
        ff2pp = FF2PP("dummy_filename")
        field = self.mock_field
        results = list(ff2pp._fields_over_all_levels(field))
        self._check_expected_levels(results, 3)

    def test__lbhem_too_small(self):
        ff2pp = FF2PP("dummy_filename")
        field = self.mock_field
        field.lbhem = 100
        with self.assertRaisesRegex(ValueError, "hence >= 101"):
            _ = list(ff2pp._fields_over_all_levels(field))

    def test__lbhem_too_large(self):
        ff2pp = FF2PP("dummy_filename")
        field = self.mock_field
        field.lbhem = 105
        with self.assertRaisesRegex(
            ValueError, "more than the total number of levels in the file = 3"
        ):
            _ = list(ff2pp._fields_over_all_levels(field))


if __name__ == "__main__":
    tests.main()



import iris
import iris.analysis
import iris.plot as iplt
import matplotlib.pyplot as plt


global_air_temp = iris.load_cube(iris.sample_data_path("air_temp.pp"))
rotated_psl = iris.load_cube(iris.sample_data_path("rotated_pole.nc"))

scheme = iris.analysis.Linear(extrapolation_mode="mask")
global_psl = rotated_psl.regrid(global_air_temp, scheme)

plt.figure(figsize=(4, 3))
iplt.pcolormesh(global_psl)
plt.title("Air pressure\n" "on a global longitude latitude grid")
ax = plt.gca()
ax.coastlines()
ax.gridlines()
ax.set_extent([-90, 70, 10, 80])

plt.show()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test function :func:`iris.util.test_file_is_newer`.

"""

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

import os
import os.path
import shutil
import tempfile

from iris.util import file_is_newer_than


class TestFileIsNewer(tests.IrisTest):
    """Test the :func:`iris.util.file_is_newer_than` function."""

    def _name2path(self, filename):
        """Add the temporary dirpath to a filename to make a full path."""
        return os.path.join(self.temp_dir, filename)

    def setUp(self):
        # make a temporary directory with testfiles of known timestamp order.
        self.temp_dir = tempfile.mkdtemp("_testfiles_tempdir")
        # define the names of some files to create
        create_file_names = [
            "older_source_1",
            "older_source_2",
            "example_result",
            "newer_source_1",
            "newer_source_2",
        ]
        # create testfiles + ensure distinct 'mtime's in the required order.
        for i_file, file_name in enumerate(create_file_names):
            file_path = self._name2path(file_name)
            with open(file_path, "w") as test_file:
                test_file.write("..content..")
            # Ensure 'mtime's are adequately separated and after create times.
            mtime = os.stat(file_path).st_mtime
            mtime += 5.0 + 10.0 * i_file
            os.utime(file_path, (mtime, mtime))

    def tearDown(self):
        # destroy whole contents of temporary directory
        shutil.rmtree(self.temp_dir)

    def _test(self, boolean_result, result_name, source_names):
        """Test expected result of executing with given args."""
        # Make args into full paths
        result_path = self._name2path(result_name)
        if isinstance(source_names, str):
            source_paths = self._name2path(source_names)
        else:
            source_paths = [self._name2path(name) for name in source_names]
        # Check result is as expected.
        self.assertEqual(
            boolean_result, file_is_newer_than(result_path, source_paths)
        )

    def test_no_sources(self):
        self._test(True, "example_result", [])

    def test_string_ok(self):
        self._test(True, "example_result", "older_source_1")

    def test_string_fail(self):
        self._test(False, "example_result", "newer_source_1")

    def test_self_result(self):
        # This fails, because same-timestamp is *not* acceptable.
        self._test(False, "example_result", "example_result")

    def test_single_ok(self):
        self._test(True, "example_result", ["older_source_2"])

    def test_single_fail(self):
        self._test(False, "example_result", ["newer_source_2"])

    def test_multiple_ok(self):
        self._test(
            True, "example_result", ["older_source_1", "older_source_2"]
        )

    def test_multiple_fail(self):
        self._test(
            False,
            "example_result",
            ["older_source_1", "older_source_2", "newer_source_1"],
        )

    def test_wild_ok(self):
        self._test(True, "example_result", ["older_sour*_*"])

    def test_wild_fail(self):
        self._test(False, "example_result", ["older_sour*", "newer_sour*"])

    def test_error_missing_result(self):
        with self.assertRaises(OSError) as error_trap:
            self._test(False, "non_exist", ["older_sour*"])
        error = error_trap.exception
        self.assertEqual(error.strerror, "No such file or directory")
        self.assertEqual(error.filename, self._name2path("non_exist"))

    def test_error_missing_source(self):
        with self.assertRaises(IOError) as error_trap:
            self._test(False, "example_result", ["older_sour*", "non_exist"])
        self.assertIn(
            "One or more of the files specified did not exist",
            str(error_trap.exception),
        )

    def test_error_missing_wild(self):
        with self.assertRaises(IOError) as error_trap:
            self._test(False, "example_result", ["older_sour*", "unknown_*"])
        self.assertIn(
            "One or more of the files specified did not exist",
            str(error_trap.exception),
        )


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test function
:func:`iris.experimental.regrid.regrid_area_weighted_rectilinear_src_and_grid`.

"""

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

import numpy as np
import numpy.ma as ma

from iris.coords import DimCoord
from iris.coord_systems import GeogCS
from iris.cube import Cube
from iris.experimental.regrid import (
    regrid_area_weighted_rectilinear_src_and_grid as regrid,
)
from iris.tests.experimental.regrid.test_regrid_area_weighted_rectilinear_src_and_grid import (
    _resampled_grid,
)


class TestMdtol(tests.IrisTest):
    # Tests to check the masking behaviour controlled by mdtol kwarg.
    def setUp(self):
        # A (3, 2, 4) cube with a masked element.
        cube = Cube(np.ma.arange(24, dtype=np.int32).reshape((3, 2, 4)))
        cs = GeogCS(6371229)
        coord = DimCoord(
            points=np.array([-1, 0, 1], dtype=np.int32),
            standard_name="latitude",
            units="degrees",
            coord_system=cs,
        )
        cube.add_dim_coord(coord, 0)
        coord = DimCoord(
            points=np.array([-1, 0, 1, 2], dtype=np.int32),
            standard_name="longitude",
            units="degrees",
            coord_system=cs,
        )
        cube.add_dim_coord(coord, 2)
        cube.coord("latitude").guess_bounds()
        cube.coord("longitude").guess_bounds()
        cube.data[1, 1, 2] = ma.masked
        self.src_cube = cube
        # Create (7, 2, 9) grid cube.
        self.grid_cube = _resampled_grid(cube, 2.3, 2.4)

    def test_default(self):
        res = regrid(self.src_cube, self.grid_cube)
        expected_mask = np.zeros((7, 2, 9), bool)
        expected_mask[2:5, 1, 4:7] = True
        self.assertArrayEqual(res.data.mask, expected_mask)

    def test_zero(self):
        res = regrid(self.src_cube, self.grid_cube, mdtol=0)
        expected_mask = np.zeros((7, 2, 9), bool)
        expected_mask[2:5, 1, 4:7] = True
        self.assertArrayEqual(res.data.mask, expected_mask)

    def test_one(self):
        res = regrid(self.src_cube, self.grid_cube, mdtol=1)
        expected_mask = np.zeros((7, 2, 9), bool)
        # Only a single cell has all contributing cells masked.
        expected_mask[3, 1, 5] = True
        self.assertArrayEqual(res.data.mask, expected_mask)

    def test_fraction_below_min(self):
        # Cells in target grid that overlap with the masked src cell
        # have the following fractions (approx. due to spherical area).
        #   4      5      6      7
        # 2 ----------------------
        #   | 0.33 | 0.66 | 0.50 |
        # 3 ----------------------
        #   | 0.33 | 1.00 | 0.75 |
        # 4 ----------------------
        #   | 0.33 | 0.66 | 0.50 |
        # 5 ----------------------
        #

        # Threshold less than minimum fraction.
        mdtol = 0.2
        res = regrid(self.src_cube, self.grid_cube, mdtol=mdtol)
        expected_mask = np.zeros((7, 2, 9), bool)
        expected_mask[2:5, 1, 4:7] = True
        self.assertArrayEqual(res.data.mask, expected_mask)

    def test_fraction_between_min_and_max(self):
        # Threshold between min and max fraction. See
        # test_fraction_below_min() comment for picture showing
        # the fractions of masked data.
        mdtol = 0.6
        res = regrid(self.src_cube, self.grid_cube, mdtol=mdtol)
        expected_mask = np.zeros((7, 2, 9), bool)
        expected_mask[2:5, 1, 5] = True
        expected_mask[3, 1, 6] = True
        self.assertArrayEqual(res.data.mask, expected_mask)

    def test_src_not_masked_array(self):
        self.src_cube.data = self.src_cube.data.filled(1.0)
        res = regrid(self.src_cube, self.grid_cube, mdtol=0.9)
        self.assertFalse(ma.isMaskedArray(res.data))

    def test_boolean_mask(self):
        self.src_cube.data = np.ma.arange(24).reshape(3, 2, 4)
        res = regrid(self.src_cube, self.grid_cube, mdtol=0.9)
        self.assertEqual(ma.count_masked(res.data), 0)

    def test_scalar_no_overlap(self):
        # Slice src so result collapses to a scalar.
        src_cube = self.src_cube[:, 1, :]
        # Regrid to a single cell with no overlap with masked src cells.
        grid_cube = self.grid_cube[2, 1, 3]
        res = regrid(src_cube, grid_cube, mdtol=0.8)
        self.assertFalse(ma.isMaskedArray(res.data))

    def test_scalar_with_overlap_below_mdtol(self):
        # Slice src so result collapses to a scalar.
        src_cube = self.src_cube[:, 1, :]
        # Regrid to a single cell with 50% overlap with masked src cells.
        grid_cube = self.grid_cube[3, 1, 4]
        # Set threshold (mdtol) to greater than 0.5 (50%).
        res = regrid(src_cube, grid_cube, mdtol=0.6)
        self.assertEqual(ma.count_masked(res.data), 0)

    def test_scalar_with_overlap_above_mdtol(self):
        # Slice src so result collapses to a scalar.
        src_cube = self.src_cube[:, 1, :]
        # Regrid to a single cell with 50% overlap with masked src cells.
        grid_cube = self.grid_cube[3, 1, 4]
        # Set threshold (mdtol) to less than 0.5 (50%).
        res = regrid(src_cube, grid_cube, mdtol=0.4)
        self.assertEqual(ma.count_masked(res.data), 1)


class TestWrapAround(tests.IrisTest):
    def test_float_tolerant_equality(self):
        # Ensure that floating point numbers are treated appropriately when
        # introducing precision difference from wrap_around.
        source = Cube([[1]])
        cs = GeogCS(6371229)

        bounds = np.array([[-91, 0]], dtype="float")
        points = bounds.mean(axis=1)
        lon_coord = DimCoord(
            points,
            bounds=bounds,
            standard_name="longitude",
            units="degrees",
            coord_system=cs,
        )
        source.add_aux_coord(lon_coord, 1)

        bounds = np.array([[-90, 90]], dtype="float")
        points = bounds.mean(axis=1)
        lat_coord = DimCoord(
            points,
            bounds=bounds,
            standard_name="latitude",
            units="degrees",
            coord_system=cs,
        )
        source.add_aux_coord(lat_coord, 0)

        grid = Cube([[0]])
        bounds = np.array([[270, 360]], dtype="float")
        points = bounds.mean(axis=1)
        lon_coord = DimCoord(
            points,
            bounds=bounds,
            standard_name="longitude",
            units="degrees",
            coord_system=cs,
        )
        grid.add_aux_coord(lon_coord, 1)
        grid.add_aux_coord(lat_coord, 0)

        res = regrid(source, grid)
        # The result should be equal to the source data and NOT be masked.
        self.assertArrayEqual(res.data, np.array([1.0]))


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for :func:`iris.fileformats.pp_load_rules._model_level_number`.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from iris.fileformats.pp_load_rules import _model_level_number


class Test_9999(tests.IrisTest):
    def test(self):
        self.assertEqual(_model_level_number(9999), 0)


class Test_lblev(tests.IrisTest):
    def test(self):
        for val in range(9999):
            self.assertEqual(_model_level_number(val), val)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the `iris.fileformats.pp._field_gen` function."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import contextlib
import io
from unittest import mock
import warnings

import numpy as np

import iris.fileformats.pp as pp


class Test(tests.IrisTest):
    @contextlib.contextmanager
    def mock_for_field_gen(self, fields):
        side_effect_fields = list(fields)[:]

        def make_pp_field_override(*args):
            # Iterates over the fields passed to this context manager,
            # until there are no more, upon which the np.fromfile
            # returns an empty list and the while loop in load() is
            # broken.
            result = side_effect_fields.pop(0)
            if not side_effect_fields:
                np.fromfile.return_value = []
            return result

        open_func = "builtins.open"
        with mock.patch("numpy.fromfile", return_value=[0]), mock.patch(
            open_func
        ), mock.patch("struct.unpack_from", return_value=[4]), mock.patch(
            "iris.fileformats.pp.make_pp_field",
            side_effect=make_pp_field_override,
        ):
            yield

    def gen_fields(self, fields):
        with self.mock_for_field_gen(fields):
            return list(pp._field_gen("mocked", "mocked"))

    def test_lblrec_invalid(self):
        pp_field = mock.Mock(lblrec=2, lbext=0)
        with warnings.catch_warnings(record=True) as warn:
            warnings.simplefilter("always")
            self.gen_fields([pp_field])
        self.assertEqual(len(warn), 1)
        wmsg = (
            "LBLREC has a different value to the .* the header in the "
            r"file \(8 and 4\)\. Skipping .*"
        )
        self.assertRegex(str(warn[0].message), wmsg)

    def test_read_headers_call(self):
        # Checks that the two calls to np.fromfile are called in the
        # expected way.
        pp_field = mock.Mock(lblrec=1, lbext=0, lbuser=[0])
        with self.mock_for_field_gen([pp_field]):
            open_fh = mock.MagicMock(spec=io.RawIOBase)
            open.return_value = open_fh
            next(pp._field_gen("mocked", read_data_bytes=False))
            with open_fh as open_fh_ctx:
                calls = [
                    mock.call(open_fh_ctx, count=45, dtype=">i4"),
                    mock.call(open_fh_ctx, count=19, dtype=">f4"),
                ]
            np.fromfile.assert_has_calls(calls)
        with open_fh as open_fh_ctx:
            expected_deferred_bytes = (
                "mocked",
                open_fh_ctx.tell(),
                4,
                np.dtype(">f4"),
            )
        self.assertEqual(pp_field.data, expected_deferred_bytes)

    def test_read_data_call(self):
        # Checks that data is read if read_data is True.
        pp_field = mock.Mock(lblrec=1, lbext=0, lbuser=[0])
        with self.mock_for_field_gen([pp_field]):
            open_fh = mock.MagicMock(spec=io.RawIOBase)
            open.return_value = open_fh
            next(pp._field_gen("mocked", read_data_bytes=True))
        with open_fh as open_fh_ctx:
            expected_loaded_bytes = pp.LoadedArrayBytes(
                open_fh_ctx.read(), np.dtype(">f4")
            )
        self.assertEqual(pp_field.data, expected_loaded_bytes)

    def test_invalid_header_release(self):
        # Check that an unknown LBREL value just results in a warning
        # and the end of the file iteration instead of raising an error.
        with self.temp_filename() as temp_path:
            np.zeros(65, dtype="i4").tofile(temp_path)
            generator = pp._field_gen(temp_path, False)
            with mock.patch("warnings.warn") as warn:
                with self.assertRaises(StopIteration):
                    next(generator)
            self.assertEqual(warn.call_count, 1)
            self.assertIn("header release number", warn.call_args[0][0])


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Routines for putting data on new strata (aka. isosurfaces), often in the
Z direction.

"""

from functools import partial

import numpy as np
import stratify

from iris.coords import Coord, AuxCoord, DimCoord
from iris.cube import Cube


def _copy_coords_without_z_dim(src, tgt, z_dim):
    """
    Helper function to copy across non z-dimenson coordinates between cubes.

    Parameters
    ----------
    src : :class:`~iris.cube.Cube`
        Incoming cube containing the coordinates to be copied from.

    tgt : :class:`~iris.cube.Cube`
        Outgoing cube for the coordinates to be copied to.

    z_dim : int
        Dimension within the `src` cube that is the z-dimension.
        This dimension will not be copied. For example, the incoming
        z-dimension cube has model level_height, whilst the outgoing
        z-dimension cube has pressure.

    """
    # Copy across non z-dimension coordinates.
    for coord in src.dim_coords:
        [dim] = src.coord_dims(coord)
        if dim != z_dim:
            tgt.add_dim_coord(coord.copy(), dim)

    for coord in src.aux_coords:
        dims = src.coord_dims(coord)
        if z_dim not in dims:
            tgt.add_aux_coord(coord.copy(), dims)

    for coord in src.derived_coords:
        dims = src.coord_dims(coord)
        if z_dim not in dims:
            tgt.add_aux_coord(coord.copy(), dims)


def relevel(cube, src_levels, tgt_levels, axis=None, interpolator=None):
    """
    Interpolate the cube onto the specified target levels, given the
    source levels of the cube.

    For example, suppose we have two datasets `P(i,j,k)` and `H(i,j,k)`
    and we want `P(i,j,H)`. We call :func:`relevel` with `cube=P`,
    `src_levels=H` and `tgt_levels` being an array of the values of `H`
    we would like.

    This routine is especially useful for computing isosurfaces of phenomenon
    that are generally monotonic in the direction of interpolation, such as
    height/pressure or salinity/depth.

    Args:

    cube : :class:`~iris.cube.Cube`
        The phenomenon data to be re-levelled.

    src_levels : :class:`~iris.cube.Cube`, :class:`~iris.coord.Coord` or string
        Describes the source levels of the `cube` that will be interpolated
        over. The `src_levels` must be in the same system as the `tgt_levels`.
        The dimensions of `src_levels` must be broadcastable to the dimensions
        of the `cube`.
        Note that, the coordinate name containing the source levels in the
        `cube` may be provided.

    tgt_levels : array-like
        Describes the target levels of the `cube` to be interpolated to. The
        `tgt_levels` must be in the same system as the `src_levels`. The
        dimensions of the `tgt_levels` must be broadcastable to the dimensions
        of the `cube`, except in the nominated axis of interpolation.

    axis : int, :class:`~iris.coords.Coord` or string
        The axis of interpolation. Defaults to the first dimension of the
        `cube`, which is typically the z-dimension. Note that, the coordinate
        name specifying the z-dimension of the `cube` may be provided.

    interpolator : callable or None
        The interpolator to use when computing the interpolation. The function
        will be passed the following positional arguments::

            (tgt-data, src-data, cube-data, axis-of-interpolation)

        If the interpolator is None, :func:`stratify.interpolate` will be used
        with linear interpolation and NaN extrapolation.

        An example of constructing an alternative interpolation scheme::

            from functools import partial
            interpolator = partial(stratify.interpolate,
                                   interpolation=stratify.INTERPOLATE_NEAREST,
                                   extrapolation=stratify.EXTRAPOLATE_LINEAR)

    """
    # Identify the z-coordinate within the phenomenon cube.
    if axis is None:
        axis = 0

    if isinstance(axis, (str, Coord)):
        [axis] = cube.coord_dims(axis)

    # Get the source level data.
    if isinstance(src_levels, str):
        src_data = cube.coord(src_levels).points
    elif isinstance(src_levels, Coord):
        src_data = src_levels.points
    else:
        src_data = src_levels.data

    # The dimensions of cube and src_data must be broadcastable.
    try:
        cube_data, src_data = np.broadcast_arrays(cube.data, src_data)
    except ValueError:
        emsg = (
            "Cannot broadcast the cube and src_levels with "
            "shapes {} and {}."
        )
        raise ValueError(emsg.format(cube.shape, src_data.shape))

    tgt_levels = np.asarray(tgt_levels)
    tgt_aux_dims = axis
    if tgt_levels.ndim != 1:
        # The dimensions of tgt_levels must be broadcastable to cube
        # in everything but the interpolation axis - otherwise raise
        # an exception.
        dim_delta = cube_data.ndim - tgt_levels.ndim
        # The axis is relative to the cube. Calculate the axis of
        # interplation relative to the tgt_levels.
        tgt_axis = axis - dim_delta
        # Calculate the cube shape without the axis of interpolation.
        data_shape = list(cube_data.shape)
        data_shape.pop(axis)
        # Calculate the tgt_levels shape without the axis of interpolation.
        target_shape = list(tgt_levels.shape)
        target_shape.pop(tgt_axis)
        # Now ensure that the shapes are broadcastable.
        try:
            np.broadcast_arrays(np.empty(data_shape), np.empty(target_shape))
        except ValueError:
            emsg = (
                "Cannot broadcast the cube and tgt_levels with "
                "shapes {} and {}, whilst ignoring axis of interpolation."
            )
            raise ValueError(emsg.format(cube_data.shape, tgt_levels.shape))
        # Calculate the dimensions over the cube that the tgt_levels span.
        tgt_aux_dims = list(range(cube_data.ndim))[dim_delta:]

    if interpolator is None:
        # Use the default stratify interpolator.
        interpolator = partial(
            stratify.interpolate, interpolation="linear", extrapolation="nan"
        )

    # Now perform the interpolation.
    new_data = interpolator(tgt_levels, src_data, cube_data, axis=axis)

    # Create a result cube with the correct shape and metadata.
    result = Cube(new_data, **cube.copy().metadata._asdict())

    # Copy across non z-dimension coordinates from the source cube
    # to the result cube.
    _copy_coords_without_z_dim(cube, result, axis)

    kwargs = dict(
        standard_name=src_levels.standard_name,
        long_name=src_levels.long_name,
        var_name=src_levels.var_name,
        units=src_levels.units,
        attributes=src_levels.attributes,
    )

    # Add our new interpolated coordinate to the result cube.
    try:
        coord = DimCoord(tgt_levels, **kwargs)
        result.add_dim_coord(coord, axis)
    except ValueError:
        # Attach the data to the trailing dimensions.
        coord = AuxCoord(tgt_levels, **kwargs)
        result.add_aux_coord(coord, tgt_aux_dims)

    return result



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the `iris.cube.CubeRepresentation` class."""

from html import escape

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from iris.coords import CellMethod, CellMeasure, AncillaryVariable
import iris.tests.stock as stock

from iris.experimental.representation import CubeRepresentation


@tests.skip_data
class Test__instantiation(tests.IrisTest):
    def setUp(self):
        self.cube = stock.simple_3d()
        self.representer = CubeRepresentation(self.cube)

    def test_cube_attributes(self):
        self.assertEqual(id(self.cube), self.representer.cube_id)
        self.assertStringEqual(str(self.cube), self.representer.cube_str)

    def test__heading_contents(self):
        content = set(self.representer.str_headings.values())
        self.assertEqual(len(content), 1)
        self.assertIsNone(list(content)[0])


@tests.skip_data
class Test__get_dim_names(tests.IrisTest):
    def setUp(self):
        self.cube = stock.realistic_4d()
        self.dim_names = [c.name() for c in self.cube.coords(dim_coords=True)]
        self.representer = CubeRepresentation(self.cube)

    def test_basic(self):
        result_names = self.representer._get_dim_names()
        self.assertEqual(result_names, self.dim_names)

    def test_one_anonymous_dim(self):
        self.cube.remove_coord("time")
        expected_names = ["--"]
        expected_names.extend(self.dim_names[1:])
        result_names = self.representer._get_dim_names()
        self.assertEqual(result_names, expected_names)

    def test_anonymous_dims(self):
        target_dims = [1, 3]
        # Replicate this here as we're about to modify it.
        expected_names = [c.name() for c in self.cube.coords(dim_coords=True)]
        for dim in target_dims:
            (this_dim_coord,) = self.cube.coords(
                contains_dimension=dim, dim_coords=True
            )
            self.cube.remove_coord(this_dim_coord)
            expected_names[dim] = "--"
        result_names = self.representer._get_dim_names()
        self.assertEqual(result_names, expected_names)


@tests.skip_data
class Test__summary_content(tests.IrisTest):
    def setUp(self):
        self.cube = stock.lat_lon_cube()
        # Check we're not tripped up by names containing spaces.
        self.cube.rename("Electron density (&<html>)")
        self.cube.units = "1e11 e/m^3"
        self.representer = CubeRepresentation(self.cube)

    def test_name(self):
        # Check the cube name is being set and formatted correctly.
        expected = escape(self.cube.name().replace("_", " ").title())
        result = self.representer.name
        self.assertEqual(expected, result)

    def test_names(self):
        # Check the dimension names used as column headings are split out and
        # formatted correctly.
        expected_coord_names = [
            c.name().replace("_", " ")
            for c in self.cube.coords(dim_coords=True)
        ]
        result_coord_names = self.representer.names[1:]
        for result in result_coord_names:
            self.assertIn(result, expected_coord_names)

    def test_units(self):
        # Check the units is being set correctly.
        expected = self.cube.units
        result = self.representer.units
        self.assertEqual(expected, result)

    def test_shapes(self):
        # Check cube dim lengths are split out correctly from the
        # summary string.
        expected = self.cube.shape
        result = self.representer.shapes
        self.assertEqual(expected, result)

    def test_ndims(self):
        expected = self.cube.ndim
        result = self.representer.ndims
        self.assertEqual(expected, result)


@tests.skip_data
class Test__get_bits(tests.IrisTest):
    def setUp(self):
        self.cube = stock.realistic_4d()
        cmth = CellMethod("mean", "time", "6hr")
        self.cube.add_cell_method(cmth)
        cms = CellMeasure([0, 1, 2, 3, 4, 5], long_name="foo")
        self.cube.add_cell_measure(cms, 0)
        avr = AncillaryVariable([0, 1, 2, 3, 4, 5], long_name="bar")
        self.cube.add_ancillary_variable(avr, 0)
        scms = CellMeasure([0], long_name="baz")
        self.cube.add_cell_measure(scms)
        self.representer = CubeRepresentation(self.cube)
        self.representer._get_bits(self.representer._get_lines())

    def test_population(self):
        for v in self.representer.str_headings.values():
            self.assertIsNotNone(v)

    def test_headings__dimcoords(self):
        contents = self.representer.str_headings["Dimension coordinates:"]
        content_str = ",".join(content for content in contents)
        dim_coords = [c.name() for c in self.cube.dim_coords]
        for coord in dim_coords:
            self.assertIn(coord, content_str)

    def test_headings__auxcoords(self):
        contents = self.representer.str_headings["Auxiliary coordinates:"]
        content_str = ",".join(content for content in contents)
        aux_coords = [
            c.name() for c in self.cube.aux_coords if c.shape != (1,)
        ]
        for coord in aux_coords:
            self.assertIn(coord, content_str)

    def test_headings__derivedcoords(self):
        contents = self.representer.str_headings["Derived coordinates:"]
        content_str = ",".join(content for content in contents)
        derived_coords = [c.name() for c in self.cube.derived_coords]
        for coord in derived_coords:
            self.assertIn(coord, content_str)

    def test_headings__cellmeasures(self):
        contents = self.representer.str_headings["Cell measures:"]
        content_str = ",".join(content for content in contents)
        cell_measures = [
            c.name() for c in self.cube.cell_measures() if c.shape != (1,)
        ]
        for coord in cell_measures:
            self.assertIn(coord, content_str)

    def test_headings__ancillaryvars(self):
        contents = self.representer.str_headings["Ancillary variables:"]
        content_str = ",".join(content for content in contents)
        ancillary_variables = [
            c.name() for c in self.cube.ancillary_variables()
        ]
        for coord in ancillary_variables:
            self.assertIn(coord, content_str)

    def test_headings__scalarcellmeasures(self):
        contents = self.representer.str_headings["Scalar cell measures:"]
        content_str = ",".join(content for content in contents)
        scalar_cell_measures = [
            c.name() for c in self.cube.cell_measures() if c.shape == (1,)
        ]
        for coord in scalar_cell_measures:
            self.assertIn(coord, content_str)

    def test_headings__scalarcoords(self):
        contents = self.representer.str_headings["Scalar coordinates:"]
        content_str = ",".join(content for content in contents)
        scalar_coords = [
            c.name() for c in self.cube.coords() if c.shape == (1,)
        ]
        for coord in scalar_coords:
            self.assertIn(coord, content_str)

    def test_headings__attributes(self):
        contents = self.representer.str_headings["Attributes:"]
        content_str = ",".join(content for content in contents)
        for attr_name, attr_value in self.cube.attributes.items():
            self.assertIn(attr_name, content_str)
            self.assertIn(attr_value, content_str)

    def test_headings__cellmethods(self):
        contents = self.representer.str_headings["Cell methods:"]
        content_str = ",".join(content for content in contents)
        for cell_method in self.cube.cell_methods:
            self.assertIn(str(cell_method), content_str)


@tests.skip_data
class Test__make_header(tests.IrisTest):
    def setUp(self):
        self.cube = stock.simple_3d()
        self.representer = CubeRepresentation(self.cube)
        self.representer._get_bits(self.representer._get_lines())
        self.header_emts = self.representer._make_header().split("\n")

    def test_name_and_units(self):
        # Check the correct name and units are being written into the top-left
        # table cell.
        # This is found in the first cell after the `<th>` is defined.
        name_and_units_cell = self.header_emts[1]
        expected = "{name} ({units})".format(
            name=self.cube.name(), units=self.cube.units
        )
        self.assertIn(expected.lower(), name_and_units_cell.lower())

    def test_number_of_columns(self):
        # There should be one headings column, plus a column per dimension.
        # Ignore opening and closing <tr> tags.
        result_cols = self.header_emts[1:-1]
        expected = self.cube.ndim + 1
        self.assertEqual(len(result_cols), expected)

    def test_row_headings(self):
        # Get only the dimension heading cells and not the headings column.
        dim_coord_names = [c.name() for c in self.cube.coords(dim_coords=True)]
        dim_col_headings = self.header_emts[2:-1]
        for coord_name, col_heading in zip(dim_coord_names, dim_col_headings):
            self.assertIn(coord_name, col_heading)


@tests.skip_data
class Test__make_shapes_row(tests.IrisTest):
    def setUp(self):
        self.cube = stock.simple_3d()
        self.representer = CubeRepresentation(self.cube)
        self.representer._get_bits(self.representer._get_lines())
        self.result = self.representer._make_shapes_row().split("\n")

    def test_row_title(self):
        title_cell = self.result[1]
        self.assertIn("Shape", title_cell)

    def test_shapes(self):
        expected_shapes = self.cube.shape
        result_shapes = self.result[2:-1]
        for expected, result in zip(expected_shapes, result_shapes):
            self.assertIn(str(expected), result)


@tests.skip_data
class Test__make_row(tests.IrisTest):
    def setUp(self):
        self.cube = stock.simple_3d()
        cm = CellMethod("mean", "time", "6hr")
        self.cube.add_cell_method(cm)
        self.representer = CubeRepresentation(self.cube)
        self.representer._get_bits(self.representer._get_lines())

    def test__title_row(self):
        title = "Wibble:"
        row = self.representer._make_row(title)
        # A cell for the title, an empty cell for each cube dimension, plus row
        # opening and closing tags.
        expected_len = self.cube.ndim + 3
        self.assertEqual(len(row), expected_len)
        # Check for specific content.
        row_str = "\n".join(element for element in row)
        self.assertIn(title.strip(":"), row_str)
        expected_html_class = "iris-title"
        self.assertIn(expected_html_class, row_str)

    def test__inclusion_row(self):
        # An inclusion row has x/- to indicate whether a coordinate describes
        # a dimension.
        title = "time"
        body = ["x", "-", "-", "-"]
        row = self.representer._make_row(title, body)
        # A cell for the title, a cell for each cube dimension, plus row
        # opening and closing tags.
        expected_len = len(body) + 3
        self.assertEqual(len(row), expected_len)
        # Check for specific content.
        row_str = "\n".join(element for element in row)
        self.assertIn(title, row_str)
        self.assertIn("x", row_str)
        self.assertIn("-", row_str)
        expected_html_class_1 = "iris-word-cell"
        expected_html_class_2 = "iris-inclusion-cell"
        self.assertIn(expected_html_class_1, row_str)
        self.assertIn(expected_html_class_2, row_str)
        # We do not expect a colspan to be set.
        self.assertNotIn("colspan", row_str)

    def test__attribute_row(self):
        # An attribute row does not contain inclusion indicators.
        title = "source"
        body = "Iris test case"
        colspan = 5
        row = self.representer._make_row(title, body, colspan)
        # We only expect two cells here: the row title cell and one other cell
        # that spans a number of columns. We also need to open and close the
        # tr html element, giving 4 bits making up the row.
        self.assertEqual(len(row), 4)
        # Check for specific content.
        row_str = "\n".join(element for element in row)
        self.assertIn(title, row_str)
        self.assertIn(body, row_str)
        # We expect a colspan to be set.
        colspan_str = 'colspan="{}"'.format(colspan)
        self.assertIn(colspan_str, row_str)


@tests.skip_data
class Test__expand_last_cell(tests.IrisTest):
    def setUp(self):
        self.cube = stock.simple_3d()
        self.representer = CubeRepresentation(self.cube)
        self.representer._get_bits(self.representer._get_lines())
        col_span = self.representer.ndims
        self.row = self.representer._make_row(
            "title", body="first", col_span=col_span
        )

    def test_add_line(self):
        cell = self.representer._expand_last_cell(self.row[-2], "second")
        self.assertIn("first<br>second", cell)


@tests.skip_data
class Test__make_content(tests.IrisTest):
    def setUp(self):
        self.cube = stock.simple_3d()
        self.representer = CubeRepresentation(self.cube)
        self.representer._get_bits(self.representer._get_lines())
        self.result = self.representer._make_content()

    def test_included(self):
        included = "Dimension coordinates"
        self.assertIn(included, self.result)
        dim_coord_names = [c.name() for c in self.cube.dim_coords]
        for coord_name in dim_coord_names:
            self.assertIn(coord_name, self.result)

    def test_not_included(self):
        # `stock.simple_3d()` only contains the `Dimension coordinates` attr.
        not_included = list(self.representer.str_headings.keys())
        not_included.pop(not_included.index("Dimension coordinates:"))
        for heading in not_included:
            self.assertNotIn(heading, self.result)

    def test_handle_newline(self):
        cube = self.cube
        cube.attributes["lines"] = "first\nsecond"
        representer = CubeRepresentation(cube)
        representer._get_bits(representer._get_lines())
        result = representer._make_content()
        self.assertIn("first<br>second", result)


@tests.skip_data
class Test_repr_html(tests.IrisTest):
    def setUp(self):
        self.cube = stock.simple_3d()
        representer = CubeRepresentation(self.cube)
        self.result = representer.repr_html()

    def test_contents_added(self):
        included = "Dimension coordinates"
        self.assertIn(included, self.result)
        not_included = "Auxiliary coordinates"
        self.assertNotIn(not_included, self.result)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Test function :func:`iris.util.array_equal`."""

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

import copy

import cf_units
import numpy as np

import iris
import iris.tests.stock as stock
from iris.util import unify_time_units


class Test(tests.IrisTest):
    def simple_1d_time_cubes(self, calendar="gregorian"):
        coord_points = [1, 2, 3, 4, 5]
        data_points = [273, 275, 278, 277, 274]
        reftimes = [
            "hours since 1970-01-01 00:00:00",
            "hours since 1970-01-02 00:00:00",
        ]
        list_of_cubes = []
        for reftime in reftimes:
            cube = iris.cube.Cube(
                np.array(data_points, dtype=np.float32),
                standard_name="air_temperature",
                units="K",
            )
            unit = cf_units.Unit(reftime, calendar=calendar)
            coord = iris.coords.DimCoord(
                points=np.array(coord_points, dtype=np.float32),
                standard_name="time",
                units=unit,
            )
            cube.add_dim_coord(coord, 0)
            list_of_cubes.append(cube)
        return list_of_cubes

    def _common(self, expected, result, coord_name="time"):
        # This tests time-like coords only.
        for cube in result:
            try:
                epoch = cube.coord(coord_name).units.origin
            except iris.exceptions.CoordinateNotFoundError:
                pass
            else:
                self.assertEqual(expected, epoch)

    def test_cubelist_with_time_coords(self):
        # Tests an :class:`iris.cube.CubeList` containing cubes with time
        # coords against a time string and a time coord.
        cubelist = iris.cube.CubeList(self.simple_1d_time_cubes())
        expected = "hours since 1970-01-01 00:00:00"
        unify_time_units(cubelist)
        self._common(expected, cubelist)

    def test_list_of_cubes_with_time_coords(self):
        # Tests an iterable containing cubes with time coords against a time
        # string and a time coord.
        list_of_cubes = self.simple_1d_time_cubes()
        expected = "hours since 1970-01-01 00:00:00"
        unify_time_units(list_of_cubes)
        self._common(expected, list_of_cubes)

    @tests.skip_data
    def test_no_time_coord_in_cubes(self):
        path0 = tests.get_data_path(("PP", "aPPglob1", "global.pp"))
        path1 = tests.get_data_path(("PP", "aPPglob1", "global_t_forecast.pp"))
        cube0 = iris.load_cube(path0)
        cube1 = iris.load_cube(path1)
        cubes = iris.cube.CubeList([cube0, cube1])
        result = copy.copy(cubes)
        unify_time_units(result)
        self.assertEqual(cubes, result)

    def test_time_coord_only_in_some_cubes(self):
        list_of_cubes = self.simple_1d_time_cubes()
        cube = stock.simple_2d()
        list_of_cubes.append(cube)
        expected = "hours since 1970-01-01 00:00:00"
        unify_time_units(list_of_cubes)
        self._common(expected, list_of_cubes)

    def test_multiple_time_coords_in_cube(self):
        cube0, cube1 = self.simple_1d_time_cubes()
        units = cf_units.Unit(
            "days since 1980-05-02 00:00:00", calendar="gregorian"
        )
        aux_coord = iris.coords.AuxCoord(
            72, standard_name="forecast_reference_time", units=units
        )
        cube1.add_aux_coord(aux_coord)
        cubelist = iris.cube.CubeList([cube0, cube1])
        expected = "hours since 1970-01-01 00:00:00"
        unify_time_units(cubelist)
        self._common(expected, cubelist)
        self._common(expected, cubelist, coord_name="forecast_reference_time")

    def test_multiple_calendars(self):
        cube0, cube1 = self.simple_1d_time_cubes()
        cube2, cube3 = self.simple_1d_time_cubes(calendar="360_day")
        cubelist = iris.cube.CubeList([cube0, cube1, cube2, cube3])
        expected = "hours since 1970-01-01 00:00:00"
        unify_time_units(cubelist)
        self._common(expected, cubelist)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

# import iris tests first so that some things can be initialised before importing anything else
import iris.tests as tests

import numpy as np

import iris.coords
from iris.coords import Cell


class TestCells(tests.IrisTest):
    def setUp(self):
        self.cell1 = iris.coords.Cell(3, [2, 4])
        self.cell2 = iris.coords.Cell(360.0, [350.0, 370.0])

    def test_cell_from_coord(self):
        Cell = iris.coords.Cell
        coord = iris.coords.AuxCoord(
            np.arange(4) * 1.5, long_name="test", units="1"
        )
        self.assertEqual(Cell(point=0.0, bound=None), coord.cell(0))
        self.assertEqual(Cell(point=1.5, bound=None), coord.cell(1))
        self.assertEqual(Cell(point=4.5, bound=None), coord.cell(-1))
        self.assertEqual(Cell(point=3.0, bound=None), coord.cell(-2))

        # put bounds on the regular coordinate
        coord.guess_bounds()
        self.assertEqual(Cell(point=0.0, bound=(-0.75, 0.75)), coord.cell(0))
        self.assertEqual(Cell(point=1.5, bound=(0.75, 2.25)), coord.cell(1))
        self.assertEqual(Cell(point=4.5, bound=(3.75, 5.25)), coord.cell(-1))
        self.assertEqual(Cell(point=3.0, bound=(2.25, 3.75)), coord.cell(-2))
        self.assertEqual(
            Cell(point=4.5, bound=(3.75, 5.25)), coord.cell(slice(-1, None))
        )

    def test_cell_from_multidim_coord(self):
        Cell = iris.coords.Cell
        coord = iris.coords.AuxCoord(
            points=np.arange(12).reshape(3, 4),
            long_name="test",
            units="1",
            bounds=np.arange(48).reshape(3, 4, 4),
        )
        self.assertRaises(IndexError, coord.cell, 0)
        self.assertEqual(
            Cell(point=3, bound=(12, 13, 14, 15)), coord.cell((0, 3))
        )

    def test_mod(self):
        # Check that applying the mod function is not modifying the original
        c = self.cell1 % 3
        self.assertNotEqual(c, self.cell1)

        c = self.cell2 % 360
        self.assertEqual(str(c), "Cell(point=0.0, bound=(350.0, 10.0))")

        c = self.cell2 % 359.13
        self.assertEqual(
            str(c),
            "Cell(point=0.8700000000000045, bound=(350.0, 10.870000000000005))",
        )

    def test_contains_point(self):
        c = iris.coords.Cell(359.5, (359.49951, 359.5004))
        self.assertTrue(c.contains_point(359.49951))

    def test_pointless(self):
        self.assertRaises(
            ValueError, iris.coords.Cell, None, (359.49951, 359.5004)
        )

    def test_add(self):
        # Check that applying the mod function is not modifying the original
        c = self.cell1 + 3
        self.assertNotEqual(c, self.cell1)

        c = self.cell2 + 360
        self.assertEqual(str(c), "Cell(point=720.0, bound=(710.0, 730.0))")

        c = self.cell2 + 359.13
        self.assertEqual(str(c), "Cell(point=719.13, bound=(709.13, 729.13))")

    def test_in(self):
        c = iris.coords.Cell(4, None)
        self.assertFalse(c in [3, 5])
        self.assertTrue(c in [3, 4])

        c = iris.coords.Cell(4, [4, 5])
        self.assertFalse(c in [3, 6])
        self.assertTrue(c in [3, 4])
        self.assertTrue(c in [3, 5])

        c = iris.coords.Cell(4, [4, 5])
        c1 = iris.coords.Cell(5, [4, 5])
        c2 = iris.coords.Cell(4, [3, 6])

        self.assertTrue(c in [3, c])
        self.assertFalse(c in [3, c1])
        self.assertFalse(c in [3, c2])

    def test_coord_equality(self):
        self.d = iris.coords.Cell(1.9, None)
        self.assertTrue(self.d == 1.9)
        self.assertFalse(self.d == [1.5, 1.9])
        self.assertFalse(self.d != 1.9)
        self.assertTrue(self.d >= 1.9)
        self.assertTrue(self.d <= 1.9)
        self.assertFalse(self.d > 1.9)
        self.assertFalse(self.d < 1.9)
        self.assertFalse(self.d in [1.5, 3.5])
        self.assertTrue(self.d in [1.5, 1.9])

        self.assertTrue(self.d != 1)
        self.assertFalse(self.d == 1)
        self.assertFalse(self.d >= 2)
        self.assertFalse(self.d <= 1)
        self.assertTrue(self.d > 1)
        self.assertTrue(self.d < 2)

        # Ensure the Cell's operators return NotImplemented.
        class Terry:
            pass

        self.assertEqual(self.d.__eq__(Terry()), NotImplemented)
        self.assertEqual(self.d.__ne__(Terry()), NotImplemented)

    def test_numpy_int_equality(self):
        dtypes = (np.int, np.int16, np.int32, np.int64)
        for dtype in dtypes:
            val = dtype(3)
            cell = iris.coords.Cell(val, None)
            self.assertEqual(cell, val)

    def test_numpy_float_equality(self):
        dtypes = (
            np.float,
            np.float16,
            np.float32,
            np.float64,
            np.float128,
            np.double,
        )
        for dtype in dtypes:
            val = dtype(3.2)
            cell = iris.coords.Cell(val, None)
            self.assertEqual(cell, val, dtype)

    def test_coord_bounds_cmp(self):
        self.e = iris.coords.Cell(0.7, [1.1, 1.9])
        self.assertTrue(self.e == 1.6)
        self.assertFalse(self.e != 1.6)
        self.assertTrue(self.e >= 1.9)
        self.assertTrue(self.e <= 1.9)
        self.assertFalse(self.e > 1.9)
        self.assertFalse(self.e < 1.9)

        self.assertFalse(self.e in [1.0, 3.5])
        self.assertTrue(self.e in [1.5, 1.9])
        self.assertTrue(self.e != 1)
        self.assertFalse(self.e == 1)
        self.assertFalse(self.e >= 2)
        self.assertFalse(self.e <= 1)
        self.assertTrue(self.e > 1)
        self.assertTrue(self.e < 2)

    def test_cell_cell_cmp(self):
        self.e = iris.coords.Cell(0.7, [1.1, 1.9])
        self.f = iris.coords.Cell(0.8, [1.1, 1.9])

        self.assertFalse(self.e > self.f)
        self.assertTrue(self.e <= self.f)
        self.assertTrue(self.f >= self.e)
        self.assertFalse(self.f < self.e)

        self.e = iris.coords.Cell(0.9, [2, 2.1])
        self.f = iris.coords.Cell(0.8, [1.1, 1.9])

        self.assertTrue(self.e > self.f)
        self.assertFalse(self.e <= self.f)
        self.assertFalse(self.f >= self.e)
        self.assertTrue(self.f < self.e)

    def test_cmp_contig(self):
        # Test cells that share an edge
        a = iris.coords.Cell(point=1054440.0, bound=(1054080.0, 1054800.0))
        b = iris.coords.Cell(point=1055160.0, bound=(1054800.0, 1055520.0))
        self.assertTrue(a < b)
        self.assertTrue(a <= b)
        self.assertFalse(a == b)
        self.assertFalse(a >= b)
        self.assertFalse(a > b)

    def test_overlap_order(self):
        # Test cells that overlap still sort correctly.
        cells = [
            Cell(point=375804.0, bound=(375792.0, 375816.0)),
            Cell(point=375672.0, bound=(375660.0, 375684.0)),
            Cell(point=375792.0, bound=(375780.0, 375804.0)),
            Cell(point=375960.0, bound=(375948.0, 375972.0)),
        ]
        sorted_cells = [
            Cell(point=375672.0, bound=(375660.0, 375684.0)),
            Cell(point=375792.0, bound=(375780.0, 375804.0)),
            Cell(point=375804.0, bound=(375792.0, 375816.0)),
            Cell(point=375960.0, bound=(375948.0, 375972.0)),
        ]
        self.assertEqual(sorted(cells), sorted_cells)

    def _check_permutations(self, a, b, a_lt_b, a_le_b, a_eq_b):
        self.assertEqual(a < b, a_lt_b)
        self.assertEqual(a <= b, a_le_b)
        self.assertEqual(a == b, a_eq_b)

        self.assertNotEqual(a > b, a_le_b)
        self.assertNotEqual(a >= b, a_lt_b)
        self.assertNotEqual(a != b, a_eq_b)

    def _check_all_permutations(self, a, b, a_lt_b, a_le_b, a_eq_b):
        self._check_permutations(a, b, a_lt_b, a_le_b, a_eq_b)
        self._check_permutations(b, a, not a_le_b, not a_lt_b, a_eq_b)

    def test_comparison_numeric(self):
        # Check what happens when you compare a simple number with a
        # point-only Cell.
        self._check_permutations(9, Cell(10), True, True, False)
        self._check_permutations(10, Cell(10), False, True, True)
        self._check_permutations(11, Cell(10), False, False, False)

    def test_comparison_numeric_with_bounds(self):
        # Check what happens when you compare a simple number with a
        # point-and-bound Cell.
        self._check_permutations(7, Cell(10, [8, 12]), True, True, False)
        self._check_permutations(8, Cell(10, [8, 12]), False, True, True)
        self._check_permutations(9, Cell(10, [8, 12]), False, True, True)
        self._check_permutations(10, Cell(10, [8, 12]), False, True, True)
        self._check_permutations(11, Cell(10, [8, 12]), False, True, True)
        self._check_permutations(12, Cell(10, [8, 12]), False, True, True)
        self._check_permutations(13, Cell(10, [8, 12]), False, False, False)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for the :class:`iris.common.lenient.Lenient`.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from unittest.mock import sentinel

from iris.common.lenient import Lenient, _LENIENT


class Test___init__(tests.IrisTest):
    def test_default(self):
        lenient = Lenient()
        expected = dict(maths=True)
        self.assertEqual(expected, lenient.__dict__)

    def test_kwargs(self):
        lenient = Lenient(maths=False)
        expected = dict(maths=False)
        self.assertEqual(expected, lenient.__dict__)

    def test_kwargs_invalid(self):
        emsg = "Invalid .* option, got 'merge'."
        with self.assertRaisesRegex(KeyError, emsg):
            _ = Lenient(merge=True)


class Test___contains__(tests.IrisTest):
    def setUp(self):
        self.lenient = Lenient()

    def test_in(self):
        self.assertTrue("maths", self.lenient)

    def test_not_in(self):
        self.assertTrue(("concatenate", self.lenient))


class Test___getitem__(tests.IrisTest):
    def setUp(self):
        self.lenient = Lenient()

    def test_in(self):
        self.assertTrue(self.lenient["maths"])

    def test_not_in(self):
        emsg = "Invalid .* option, got 'MATHS'."
        with self.assertRaisesRegex(KeyError, emsg):
            _ = self.lenient["MATHS"]


class Test___repr__(tests.IrisTest):
    def setUp(self):
        self.lenient = Lenient()

    def test(self):
        expected = "Lenient(maths=True)"
        self.assertEqual(expected, repr(self.lenient))


class Test___setitem__(tests.IrisTest):
    def setUp(self):
        self.lenient = Lenient()

    def test_key_invalid(self):
        emsg = "Invalid .* option, got 'MATHS."
        with self.assertRaisesRegex(KeyError, emsg):
            self.lenient["MATHS"] = False

    def test_maths_value_invalid(self):
        value = sentinel.value
        emsg = f"Invalid .* option 'maths' value, got {value!r}."
        with self.assertRaisesRegex(ValueError, emsg):
            self.lenient["maths"] = value

    def test_maths_disable__lenient_enable_true(self):
        self.assertTrue(_LENIENT.enable)
        self.lenient["maths"] = False
        self.assertFalse(self.lenient.__dict__["maths"])
        self.assertFalse(_LENIENT.enable)

    def test_maths_disable__lenient_enable_false(self):
        _LENIENT.__dict__["enable"] = False
        self.assertFalse(_LENIENT.enable)
        self.lenient["maths"] = False
        self.assertFalse(self.lenient.__dict__["maths"])
        self.assertFalse(_LENIENT.enable)

    def test_maths_enable__lenient_enable_true(self):
        self.assertTrue(_LENIENT.enable)
        self.lenient["maths"] = True
        self.assertTrue(self.lenient.__dict__["maths"])
        self.assertTrue(_LENIENT.enable)

    def test_maths_enable__lenient_enable_false(self):
        _LENIENT.__dict__["enable"] = False
        self.assertFalse(_LENIENT.enable)
        self.lenient["maths"] = True
        self.assertTrue(self.lenient.__dict__["maths"])
        self.assertTrue(_LENIENT.enable)


class Test_context(tests.IrisTest):
    def setUp(self):
        self.lenient = Lenient()

    def test_nop(self):
        self.assertTrue(self.lenient["maths"])

        with self.lenient.context():
            self.assertTrue(self.lenient["maths"])

        self.assertTrue(self.lenient["maths"])

    def test_maths_disable__lenient_true(self):
        # synchronised
        self.assertTrue(_LENIENT.enable)
        self.assertTrue(self.lenient["maths"])

        with self.lenient.context(maths=False):
            # still synchronised
            self.assertFalse(_LENIENT.enable)
            self.assertFalse(self.lenient["maths"])

        # still synchronised
        self.assertTrue(_LENIENT.enable)
        self.assertTrue(self.lenient["maths"])

    def test_maths_disable__lenient_false(self):
        # not synchronised
        _LENIENT.__dict__["enable"] = False
        self.assertFalse(_LENIENT.enable)
        self.assertTrue(self.lenient["maths"])

        with self.lenient.context(maths=False):
            # now synchronised
            self.assertFalse(_LENIENT.enable)
            self.assertFalse(self.lenient["maths"])

        # still synchronised
        self.assertTrue(_LENIENT.enable)
        self.assertTrue(self.lenient["maths"])

    def test_maths_enable__lenient_true(self):
        # not synchronised
        self.assertTrue(_LENIENT.enable)
        self.lenient.__dict__["maths"] = False
        self.assertFalse(self.lenient["maths"])

        with self.lenient.context(maths=True):
            # now synchronised
            self.assertTrue(_LENIENT.enable)
            self.assertTrue(self.lenient["maths"])

        # still synchronised
        self.assertFalse(_LENIENT.enable)
        self.assertFalse(self.lenient["maths"])

    def test_maths_enable__lenient_false(self):
        # synchronised
        _LENIENT.__dict__["enable"] = False
        self.assertFalse(_LENIENT.enable)
        self.lenient.__dict__["maths"] = False
        self.assertFalse(self.lenient["maths"])

        with self.lenient.context(maths=True):
            # still synchronised
            self.assertTrue(_LENIENT.enable)
            self.assertTrue(self.lenient["maths"])

        # still synchronised
        self.assertFalse(_LENIENT.enable)
        self.assertFalse(self.lenient["maths"])



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test iris.fileformats.rules.py - metadata translation rules.

"""

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

import types
from unittest import mock

import numpy as np

from iris.aux_factory import HybridHeightFactory
from iris.cube import Cube
from iris.fileformats.rules import (
    ConcreteReferenceTarget,
    ConversionMetadata,
    Factory,
    Loader,
    Reference,
    ReferenceTarget,
    load_cubes,
    scalar_cell_method,
)
from iris.coords import CellMethod
import iris.tests.stock as stock


class TestConcreteReferenceTarget(tests.IrisTest):
    def test_attributes(self):
        with self.assertRaises(TypeError):
            target = ConcreteReferenceTarget()

        target = ConcreteReferenceTarget("foo")
        self.assertEqual(target.name, "foo")
        self.assertIsNone(target.transform)

        def transform(_):
            return _

        target = ConcreteReferenceTarget("foo", transform)
        self.assertEqual(target.name, "foo")
        self.assertIs(target.transform, transform)

    def test_single_cube_no_transform(self):
        target = ConcreteReferenceTarget("foo")
        src = stock.simple_2d()
        target.add_cube(src)
        self.assertIs(target.as_cube(), src)

    def test_single_cube_with_transform(self):
        def transform(cube):
            return {"long_name": "wibble"}

        target = ConcreteReferenceTarget("foo", transform)
        src = stock.simple_2d()
        target.add_cube(src)
        dest = target.as_cube()
        self.assertEqual(dest.long_name, "wibble")
        self.assertNotEqual(dest, src)
        dest.long_name = src.long_name
        self.assertEqual(dest, src)

    @tests.skip_data
    def test_multiple_cubes_no_transform(self):
        target = ConcreteReferenceTarget("foo")
        src = stock.realistic_4d()
        for i in range(src.shape[0]):
            target.add_cube(src[i])
        dest = target.as_cube()
        self.assertIsNot(dest, src)
        self.assertEqual(dest, src)

    @tests.skip_data
    def test_multiple_cubes_with_transform(self):
        def transform(cube):
            return {"long_name": "wibble"}

        target = ConcreteReferenceTarget("foo", transform)
        src = stock.realistic_4d()
        for i in range(src.shape[0]):
            target.add_cube(src[i])
        dest = target.as_cube()
        self.assertEqual(dest.long_name, "wibble")
        self.assertNotEqual(dest, src)
        dest.long_name = src.long_name
        self.assertEqual(dest, src)


class TestLoadCubes(tests.IrisTest):
    def test_simple_factory(self):
        # Test the creation process for a factory definition which only
        # uses simple dict arguments.

        # Make a minimal fake data object that passes as lazy data.
        core_data_array = mock.Mock(compute=None, dtype=np.dtype("f4"))
        # Make a fake PPField which will be supplied to our converter.
        field = mock.Mock(
            core_data=mock.Mock(return_value=core_data_array),
            realised_dtype=np.dtype("f4"),
            bmdi=None,
        )

        def field_generator(filename):
            return [field]

        # A fake conversion function returning:
        #   1) A parameter cube needing a simple factory construction.
        aux_factory = mock.Mock()
        factory = mock.Mock()
        factory.args = [{"name": "foo"}]
        factory.factory_class = (
            lambda *args: setattr(aux_factory, "fake_args", args)
            or aux_factory
        )

        def converter(field):
            return ConversionMetadata(
                [factory], [], "", "", "", {}, [], [], []
            )

        # Finish by making a fake Loader
        fake_loader = Loader(field_generator, {}, converter)
        cubes = load_cubes(["fake_filename"], None, fake_loader)

        # Check the result is a generator with a single entry.
        self.assertIsInstance(cubes, types.GeneratorType)
        try:
            # Suppress the normal Cube.coord() and Cube.add_aux_factory()
            # methods.
            coord_method = Cube.coord
            add_aux_factory_method = Cube.add_aux_factory
            Cube.coord = lambda self, **args: args
            Cube.add_aux_factory = lambda self, aux_factory: setattr(
                self, "fake_aux_factory", aux_factory
            )

            cubes = list(cubes)
        finally:
            Cube.coord = coord_method
            Cube.add_aux_factory = add_aux_factory_method
        self.assertEqual(len(cubes), 1)
        # Check the "cube" has an "aux_factory" added, which itself
        # must have been created with the correct arguments.
        self.assertTrue(hasattr(cubes[0], "fake_aux_factory"))
        self.assertIs(cubes[0].fake_aux_factory, aux_factory)
        self.assertTrue(hasattr(aux_factory, "fake_args"))
        self.assertEqual(aux_factory.fake_args, ({"name": "foo"},))

    @tests.skip_data
    def test_cross_reference(self):
        # Test the creation process for a factory definition which uses
        # a cross-reference.

        param_cube = stock.realistic_4d_no_derived()
        orog_coord = param_cube.coord("surface_altitude")
        param_cube.remove_coord(orog_coord)

        orog_cube = param_cube[0, 0, :, :]
        orog_cube.data = orog_coord.points
        orog_cube.rename("surface_altitude")
        orog_cube.units = orog_coord.units
        orog_cube.attributes = orog_coord.attributes

        # We're going to test for the presence of the hybrid height
        # stuff later, so let's make sure it's not already there!
        assert len(param_cube.aux_factories) == 0
        assert not param_cube.coords("surface_altitude")

        # The fake PPFields which will be supplied to our converter.
        press_field = mock.Mock(
            core_data=mock.Mock(return_value=param_cube.data),
            bmdi=-1e20,
            realised_dtype=param_cube.dtype,
        )

        orog_field = mock.Mock(
            core_data=mock.Mock(return_value=orog_cube.data),
            bmdi=-1e20,
            realised_dtype=orog_cube.dtype,
        )

        def field_generator(filename):
            return [press_field, orog_field]

        # A fake rule set returning:
        #   1) A parameter cube needing an "orography" reference
        #   2) An "orography" cube

        def converter(field):
            if field is press_field:
                src = param_cube
                factories = [
                    Factory(HybridHeightFactory, [Reference("orography")])
                ]
                references = []
            else:
                src = orog_cube
                factories = []
                references = [ReferenceTarget("orography", None)]
            dim_coords_and_dims = [
                (coord, src.coord_dims(coord)[0]) for coord in src.dim_coords
            ]
            aux_coords_and_dims = [
                (coord, src.coord_dims(coord)) for coord in src.aux_coords
            ]
            return ConversionMetadata(
                factories,
                references,
                src.standard_name,
                src.long_name,
                src.units,
                src.attributes,
                src.cell_methods,
                dim_coords_and_dims,
                aux_coords_and_dims,
            )

        # Finish by making a fake Loader
        fake_loader = Loader(field_generator, {}, converter)
        cubes = load_cubes(["fake_filename"], None, fake_loader)

        # Check the result is a generator containing two Cubes.
        self.assertIsInstance(cubes, types.GeneratorType)
        cubes = list(cubes)
        self.assertEqual(len(cubes), 2)
        # Check the "cube" has an "aux_factory" added, which itself
        # must have been created with the correct arguments.
        self.assertEqual(len(cubes[1].aux_factories), 1)
        self.assertEqual(len(cubes[1].coords("surface_altitude")), 1)


class Test_scalar_cell_method(tests.IrisTest):
    """Tests for iris.fileformats.rules.scalar_cell_method() function"""

    def setUp(self):
        self.cube = stock.simple_2d()
        self.cm = CellMethod("mean", "foo", "1 hour")
        self.cube.cell_methods = (self.cm,)

    def test_cell_method_found(self):
        actual = scalar_cell_method(self.cube, "mean", "foo")
        self.assertEqual(actual, self.cm)

    def test_method_different(self):
        actual = scalar_cell_method(self.cube, "average", "foo")
        self.assertIsNone(actual)

    def test_coord_name_different(self):
        actual = scalar_cell_method(self.cube, "average", "bar")
        self.assertIsNone(actual)

    def test_double_coord_fails(self):
        self.cube.cell_methods = (
            CellMethod("mean", ("foo", "bar"), ("1 hour", "1 hour")),
        )
        actual = scalar_cell_method(self.cube, "mean", "foo")
        self.assertIsNone(actual)


if __name__ == "__main__":
    tests.main()



"""
Tri-Polar Grid Projected Plotting
=================================

This example demonstrates cell plots of data on the semi-structured ORCA2 model
grid.

First, the data is projected into the PlateCarree coordinate reference system.

Second four pcolormesh plots are created from this projected dataset,
using different projections for the output image.

"""

import cartopy.crs as ccrs
import matplotlib.pyplot as plt

import iris
import iris.analysis.cartography
import iris.plot as iplt
import iris.quickplot as qplt


def main():
    # Load data
    filepath = iris.sample_data_path("orca2_votemper.nc")
    cube = iris.load_cube(filepath)

    # Choose plot projections
    projections = {}
    projections["Mollweide"] = ccrs.Mollweide()
    projections["PlateCarree"] = ccrs.PlateCarree()
    projections["NorthPolarStereo"] = ccrs.NorthPolarStereo()
    projections["Orthographic"] = ccrs.Orthographic(
        central_longitude=-90, central_latitude=45
    )

    pcarree = projections["PlateCarree"]
    # Transform cube to target projection
    new_cube, extent = iris.analysis.cartography.project(
        cube, pcarree, nx=400, ny=200
    )

    # Plot data in each projection
    for name in sorted(projections):
        fig = plt.figure()
        fig.suptitle("ORCA2 Data Projected to {}".format(name))
        # Set up axes and title
        ax = plt.subplot(projection=projections[name])
        # Set limits
        ax.set_global()
        # plot with Iris quickplot pcolormesh
        qplt.pcolormesh(new_cube)
        # Draw coastlines
        ax.coastlines()

        iplt.show()


if __name__ == "__main__":
    main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for the :class:`iris.common.metadata._NamedTupleMeta`.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from abc import abstractmethod

from iris.common.metadata import _NamedTupleMeta


class Test(tests.IrisTest):
    @staticmethod
    def names(classes):
        return [cls.__name__ for cls in classes]

    @staticmethod
    def emsg_generate(members):
        if isinstance(members, str):
            members = (members,)
        emsg = ".* missing {} required positional argument{}: {}"
        args = ", ".join([f"{member!r}" for member in members[:-1]])
        count = len(members)
        if count == 1:
            args += f"{members[-1]!r}"
        elif count == 2:
            args += f" and {members[-1]!r}"
        else:
            args += f", and {members[-1]!r}"
        plural = "s" if count > 1 else ""
        return emsg.format(len(members), plural, args)

    def test__no_bases_with_abstract_members_property(self):
        class Metadata(metaclass=_NamedTupleMeta):
            @property
            @abstractmethod
            def _members(self):
                pass

        expected = ["object"]
        self.assertEqual(self.names(Metadata.__bases__), expected)
        expected = ["Metadata", "object"]
        self.assertEqual(self.names(Metadata.__mro__), expected)
        emsg = (
            "Can't instantiate abstract class .* with abstract "
            "methods _members"
        )
        with self.assertRaisesRegex(TypeError, emsg):
            _ = Metadata()

    def test__no_bases_single_member(self):
        member = "arg_one"

        class Metadata(metaclass=_NamedTupleMeta):
            _members = member

        expected = ["MetadataNamedtuple"]
        self.assertEqual(self.names(Metadata.__bases__), expected)
        expected = ["Metadata", "MetadataNamedtuple", "tuple", "object"]
        self.assertEqual(self.names(Metadata.__mro__), expected)
        emsg = self.emsg_generate(member)
        with self.assertRaisesRegex(TypeError, emsg):
            _ = Metadata()
        metadata = Metadata(1)
        self.assertEqual(metadata._fields, (member,))
        self.assertEqual(metadata.arg_one, 1)

    def test__no_bases_multiple_members(self):
        members = ("arg_one", "arg_two")

        class Metadata(metaclass=_NamedTupleMeta):
            _members = members

        expected = ["MetadataNamedtuple"]
        self.assertEqual(self.names(Metadata.__bases__), expected)
        expected = ["Metadata", "MetadataNamedtuple", "tuple", "object"]
        self.assertEqual(self.names(Metadata.__mro__), expected)
        emsg = self.emsg_generate(members)
        with self.assertRaisesRegex(TypeError, emsg):
            _ = Metadata()
        values = range(len(members))
        metadata = Metadata(*values)
        self.assertEqual(metadata._fields, members)
        expected = dict(zip(members, values))
        self.assertEqual(metadata._asdict(), expected)

    def test__multiple_bases_multiple_members(self):
        members_parent = ("arg_one", "arg_two")
        members_child = ("arg_three", "arg_four")

        class MetadataParent(metaclass=_NamedTupleMeta):
            _members = members_parent

        class MetadataChild(MetadataParent):
            _members = members_child

        # Check the parent class...
        expected = ["MetadataParentNamedtuple"]
        self.assertEqual(self.names(MetadataParent.__bases__), expected)
        expected = [
            "MetadataParent",
            "MetadataParentNamedtuple",
            "tuple",
            "object",
        ]
        self.assertEqual(self.names(MetadataParent.__mro__), expected)
        emsg = self.emsg_generate(members_parent)
        with self.assertRaisesRegex(TypeError, emsg):
            _ = MetadataParent()
        values_parent = range(len(members_parent))
        metadata_parent = MetadataParent(*values_parent)
        self.assertEqual(metadata_parent._fields, members_parent)
        expected = dict(zip(members_parent, values_parent))
        self.assertEqual(metadata_parent._asdict(), expected)

        # Check the dependant child class...
        expected = ["MetadataChildNamedtuple", "MetadataParent"]
        self.assertEqual(self.names(MetadataChild.__bases__), expected)
        expected = [
            "MetadataChild",
            "MetadataChildNamedtuple",
            "MetadataParent",
            "MetadataParentNamedtuple",
            "tuple",
            "object",
        ]
        self.assertEqual(self.names(MetadataChild.__mro__), expected)
        emsg = self.emsg_generate((*members_parent, *members_child))
        with self.assertRaisesRegex(TypeError, emsg):
            _ = MetadataChild()
        fields_child = (*members_parent, *members_child)
        values_child = range(len(fields_child))
        metadata_child = MetadataChild(*values_child)
        self.assertEqual(metadata_child._fields, fields_child)
        expected = dict(zip(fields_child, values_child))
        self.assertEqual(metadata_child._asdict(), expected)


if __name__ == "__main__":
    tests.main()
